{ ******************************************************* }
{ }
{ Delphi Visual Component Library }
{ }
{ Copyright(c) 1995-2010 Embarcadero Technologies, Inc. }
{ }
{ Message-digest algorithm based on C sample }
{ provided in Appendix section of }
{ Rivest-MD5.TXT available at: }
{ http://theory.lcs.mit.edu/~rivest/Rivest-MD5.txt }
{ }
{ ******************************************************* }
unit rMessageDigest;

interface

uses Types;

type

  IMD5 = interface
    ['{887C9EF0-15B9-41B4-A403-0431793B6E41}']
    procedure Init;
    procedure Update(const Input: TByteDynArray; Len: Longword); overload;
    procedure Update(const Input: string); overload;
    function Final: TByteDynArray;
    function AsString: string;
    function AsGUID: string;
  end;

function GetMD5: IMD5;

implementation

{$Q-}
{$R-}

const
  S11 = 7;
  S12 = 12;
  S13 = 17;
  S14 = 22;

  S21 = 5;
  S22 = 9;
  S23 = 14;
  S24 = 20;

  S31 = 4;
  S32 = 11;
  S33 = 16;
  S34 = 23;

  S41 = 6;
  S42 = 10;
  S43 = 15;
  S44 = 21;

var
  padding: TByteDynArray;

type

  ArrayOfLWord = array of Longword;

  TMD5 = class(TInterfacedObject, IMD5)
  private
    FContextState: ArrayOfLWord;
    FContextCount: ArrayOfLWord;
    FContextBuffer: TByteDynArray;
  public
    constructor Create;
    procedure Init; virtual;
    procedure Update(const Input: TByteDynArray; Len: Longword); overload; virtual;
    procedure Update(const Input: string); overload; virtual;
    function Final: TByteDynArray; virtual;
    procedure Transform(const block: TByteDynArray; shift: Integer);
    procedure Decode(var Dst: ArrayOfLWord; const Src: TByteDynArray; Len: Integer; shift: Integer);
    procedure Encode(var Dst: TByteDynArray; const Src: ArrayOfLWord; Len: Integer);
    function AsString: string;
    function AsGUID: string;
  end;

function GetMD5: IMD5;
begin
  Result:=TMD5.Create;
end;

constructor TMD5.Create;
begin
  inherited Create;
  Init;
end;

procedure TMD5.Init;
begin
  SetLength(FContextCount, 2);
  FContextCount[0]:=0;
  FContextCount[1]:=0;
  SetLength(FContextState, 4);
  FContextState[0]:=$67452301;
  FContextState[1]:=$EFCDAB89;
  FContextState[2]:=$98BADCFE;
  FContextState[3]:=$10325476;
  SetLength(FContextBuffer, 64);
end;

procedure TMD5.Update(const Input: string);
var
{$IFDEF UNICODE}
  utf8Str: UTF8String;
{$ENDIF}
  Bytes: TByteDynArray;
  Len: Integer;
  Str: PAnsiChar;
begin
{$IFDEF UNICODE}
  utf8Str:=UTF8Encode(Input);
  Len:=Length(utf8Str);
{$ELSE}
  Len:=Length(Input);
{$ENDIF}
  if Len > 0 then
  begin
    SetLength(Bytes, Len);
{$IFDEF UNICODE}
    Str:=PAnsiChar(utf8Str);
{$ELSE}
    Str:=PAnsiChar(Input);
{$ENDIF}
    Move(Str^, Pointer(Bytes)^, Len);
    Update(Bytes, Len);
  end;
end;

procedure TMD5.Update(const Input: TByteDynArray; Len: Longword);
var
  index, partlen, I, start: Longword;
begin
  { Compute number of bytes mod 64 }
  index:=(FContextCount[0] shr 3) and $3F;
  { Update number of bits }
  Inc(FContextCount[0], Len shl 3);
  if (FContextCount[0] < (Len shl 3)) then
    Inc(FContextCount[1]);
  Inc(FContextCount[1], Len shr 29);
  partlen:=64 - index;

  { Transform (as many times as possible) }
  if Len >= partlen then
  begin
    for I:=0 to partlen - 1 do
      FContextBuffer[I + index]:=Input[I];

    Transform(FContextBuffer, 0);
    I:=partlen;
    while (I + 63) < Len do
    begin
      Transform(Input, I);
      Inc(I, 64);
    end;
    index:=0;
  end
  else
    I:=0;

  { Input remaining input }
  if (I < Len) then
  begin
    start:=I;
    while (I < Len) do
    begin
      FContextBuffer[index + I - start]:=Input[I];
      Inc(I);
    end;
  end;
end;

function TMD5.Final: TByteDynArray;
var
  bits: TByteDynArray;
  index, padlen: Integer;
begin
  { Save number of bits }
  Encode(bits, FContextCount, 8);
  { Pad out to 56 mod 64 }
  index:=(FContextCount[0] shr 3) and $3F;
  if index < 56 then
    padlen:=56 - index
  else
    padlen:=120 - index;

  Update(padding, padlen);
  { Append length (before padding) }
  Update(bits, 8);
  { Store state in digest }
  Encode(Result, FContextState, 16);
end;

function TMD5.AsString: string;
const
  XD: array [0 .. 15] of char = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F');
var
  digest: TByteDynArray;
  I: Integer;
begin
  Result:='';
  digest:=Final;
  for I:=0 to Length(digest) - 1 do
    Result:=Result + XD[(digest[I] shr 4) and $0F] + XD[digest[I] and $0F];
end;

function TMD5.AsGUID: string;
const
  XD: array [0 .. 15] of char = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F');
var
  digest: TByteDynArray;
  I: Integer;
begin
  Result:='';
  digest:=Final;
  for I:=0 to Length(digest) - 1 do
  begin
    case I of
      0:
        Result:=Result + '{';
      4, 6, 8, 10:
        Result:=Result + '-';
    end;
    Result:=Result + XD[(digest[I] shr 4) and $0F] + XD[digest[I] and $0F];
    if I = 15 then
      Result:=Result + '}';
  end;
end;

procedure TMD5.Transform(const block: TByteDynArray; shift: Integer);

  function F(x, y, z: Longword): Longword;
  begin
    Result:=(x and y) or ((not x) and z);
  end;

  function G(x, y, z: Longword): Longword;
  begin
    Result:=(x and z) or (y and (not z));
  end;

  function H(x, y, z: Longword): Longword;
  begin
    Result:=x xor y xor z;
  end;

  function I(x, y, z: Longword): Longword;
  begin
    Result:=y xor (x or (not z));
  end;

  procedure RL(var x: Longword; n: Byte);
  begin
    x:=(x shl n) or (x shr (32 - n));
  end;

  procedure FF(var a: Longword; b, c, d, x: Longword; s: Byte; ac: Longword);
  begin
    Inc(a, F(b, c, d) + x + ac);
    RL(a, s);
    Inc(a, b);
  end;

  procedure GG(var a: Longword; b, c, d, x: Longword; s: Byte; ac: Longword);
  begin
    Inc(a, G(b, c, d) + x + ac);
    RL(a, s);
    Inc(a, b);
  end;

  procedure HH(var a: Longword; b, c, d, x: Longword; s: Byte; ac: Longword);
  begin
    Inc(a, H(b, c, d) + x + ac);
    RL(a, s);
    Inc(a, b);
  end;

  procedure II(var a: Longword; b, c, d, x: Longword; s: Byte; ac: Longword);
  begin
    Inc(a, I(b, c, d) + x + ac);
    RL(a, s);
    Inc(a, b);
  end;

var
  a, b, c, d: Longword;
  x: ArrayOfLWord;

begin
  a:=FContextState[0];
  b:=FContextState[1];
  c:=FContextState[2];
  d:=FContextState[3];

  Decode(x, block, 64, shift);

  { Round 1 }
  FF(a, b, c, d, x[0], S11, $D76AA478); { 1 }
  FF(d, a, b, c, x[1], S12, $E8C7B756); { 2 }
  FF(c, d, a, b, x[2], S13, $242070DB); { 3 }
  FF(b, c, d, a, x[3], S14, $C1BDCEEE); { 4 }
  FF(a, b, c, d, x[4], S11, $F57C0FAF); { 5 }
  FF(d, a, b, c, x[5], S12, $4787C62A); { 6 }
  FF(c, d, a, b, x[6], S13, $A8304613); { 7 }
  FF(b, c, d, a, x[7], S14, $FD469501); { 8 }
  FF(a, b, c, d, x[8], S11, $698098D8); { 9 }
  FF(d, a, b, c, x[9], S12, $8B44F7AF); { 10 }
  FF(c, d, a, b, x[10], S13, $FFFF5BB1); { 11 }
  FF(b, c, d, a, x[11], S14, $895CD7BE); { 12 }
  FF(a, b, c, d, x[12], S11, $6B901122); { 13 }
  FF(d, a, b, c, x[13], S12, $FD987193); { 14 }
  FF(c, d, a, b, x[14], S13, $A679438E); { 15 }
  FF(b, c, d, a, x[15], S14, $49B40821); { 16 }

  { Round 2 }
  GG(a, b, c, d, x[1], S21, $F61E2562); { 17 }
  GG(d, a, b, c, x[6], S22, $C040B340); { 18 }
  GG(c, d, a, b, x[11], S23, $265E5A51); { 19 }
  GG(b, c, d, a, x[0], S24, $E9B6C7AA); { 20 }
  GG(a, b, c, d, x[5], S21, $D62F105D); { 21 }
  GG(d, a, b, c, x[10], S22, $2441453); { 22 }
  GG(c, d, a, b, x[15], S23, $D8A1E681); { 23 }
  GG(b, c, d, a, x[4], S24, $E7D3FBC8); { 24 }
  GG(a, b, c, d, x[9], S21, $21E1CDE6); { 25 }
  GG(d, a, b, c, x[14], S22, $C33707D6); { 26 }
  GG(c, d, a, b, x[3], S23, $F4D50D87); { 27 }
  GG(b, c, d, a, x[8], S24, $455A14ED); { 28 }
  GG(a, b, c, d, x[13], S21, $A9E3E905); { 29 }
  GG(d, a, b, c, x[2], S22, $FCEFA3F8); { 30 }
  GG(c, d, a, b, x[7], S23, $676F02D9); { 31 }
  GG(b, c, d, a, x[12], S24, $8D2A4C8A); { 32 }

  { Round 3 }
  HH(a, b, c, d, x[5], S31, $FFFA3942); { 33 }
  HH(d, a, b, c, x[8], S32, $8771F681); { 34 }
  HH(c, d, a, b, x[11], S33, $6D9D6122); { 35 }
  HH(b, c, d, a, x[14], S34, $FDE5380C); { 36 }
  HH(a, b, c, d, x[1], S31, $A4BEEA44); { 37 }
  HH(d, a, b, c, x[4], S32, $4BDECFA9); { 38 }
  HH(c, d, a, b, x[7], S33, $F6BB4B60); { 39 }
  HH(b, c, d, a, x[10], S34, $BEBFBC70); { 40 }
  HH(a, b, c, d, x[13], S31, $289B7EC6); { 41 }
  HH(d, a, b, c, x[0], S32, $EAA127FA); { 42 }
  HH(c, d, a, b, x[3], S33, $D4EF3085); { 43 }
  HH(b, c, d, a, x[6], S34, $4881D05); { 44 }
  HH(a, b, c, d, x[9], S31, $D9D4D039); { 45 }
  HH(d, a, b, c, x[12], S32, $E6DB99E5); { 46 }
  HH(c, d, a, b, x[15], S33, $1FA27CF8); { 47 }
  HH(b, c, d, a, x[2], S34, $C4AC5665); { 48 }

  { Round 4 }
  II(a, b, c, d, x[0], S41, $F4292244); { 49 }
  II(d, a, b, c, x[7], S42, $432AFF97); { 50 }
  II(c, d, a, b, x[14], S43, $AB9423A7); { 51 }
  II(b, c, d, a, x[5], S44, $FC93A039); { 52 }
  II(a, b, c, d, x[12], S41, $655B59C3); { 53 }
  II(d, a, b, c, x[3], S42, $8F0CCC92); { 54 }
  II(c, d, a, b, x[10], S43, $FFEFF47D); { 55 }
  II(b, c, d, a, x[1], S44, $85845DD1); { 56 }
  II(a, b, c, d, x[8], S41, $6FA87E4F); { 57 }
  II(d, a, b, c, x[15], S42, $FE2CE6E0); { 58 }
  II(c, d, a, b, x[6], S43, $A3014314); { 59 }
  II(b, c, d, a, x[13], S44, $4E0811A1); { 60 }
  II(a, b, c, d, x[4], S41, $F7537E82); { 61 }
  II(d, a, b, c, x[11], S42, $BD3AF235); { 62 }
  II(c, d, a, b, x[2], S43, $2AD7D2BB); { 63 }
  II(b, c, d, a, x[9], S44, $EB86D391); { 64 }

  Inc(FContextState[0], a);
  Inc(FContextState[1], b);
  Inc(FContextState[2], c);
  Inc(FContextState[3], d);
end;

procedure TMD5.Encode(var Dst: TByteDynArray; const Src: ArrayOfLWord; Len: Integer);
var
  I, j: Integer;
begin
  I:=0;
  j:=0;
  SetLength(Dst, Len);
  while (j < Len) do
  begin
    Dst[j]:=Byte((Src[I] and $FF));
    Dst[j + 1]:=Byte((Src[I] shr 8) and $FF);
    Dst[j + 2]:=Byte((Src[I] shr 16) and $FF);
    Dst[j + 3]:=Byte((Src[I] shr 24) and $FF);
    Inc(j, 4);
    Inc(I);
  end;
end;

procedure TMD5.Decode(var Dst: ArrayOfLWord; const Src: TByteDynArray; Len: Integer; shift: Integer);
var
  I, j: Integer;
  a, b, c, d: Byte;
begin
  j:=0;
  I:=0;
  SetLength(Dst, 16);
  while (j < Len) do
  begin
    a:=Src[j + shift];
    b:=Src[j + shift + 1];
    c:=Src[j + shift + 2];
    d:=Src[j + shift + 3];
    Dst[I]:=Longword(a and $FF) or (Longword(b and $FF) shl 8) or (Longword(c and $FF) shl 16) or (Longword(d and $FF) shl 24);
    Inc(j, 4);
    Inc(I);
  end;
end;

initialization

SetLength(padding, 64);
padding[0]:=$80;

{
  MD5 test suite:
  MD5 ("") = d41d8cd98f00b204e9800998ecf8427e
  MD5 ("a") = 0cc175b9c0f1b6a831c399e269772661
  MD5 ("abc") = 900150983cd24fb0d6963f7d28e17f72
  MD5 ("message digest") = f96b697d7cb7938d525a2f31aaf161d0
  MD5 ("abcdefghijklmnopqrstuvwxyz") = c3fcd3d76192e4007dfb496cca67e13b
  MD5 ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789") =
  d174ab98d277d9f5a5611c2c9f419d9f
  MD5 ("123456789012345678901234567890123456789012345678901234567890123456
  78901234567890") = 57edf4a22be3c955ac49da2e2107b67a
}

end.
